"""
Name: MCS_Classifier.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA 
Note: Code primarily based on python XCS authored by Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
Contact: ryan.j.urbanowicz@darmouth.edu
Created: 3/20/2009
Updated: 10/15/2013
Description: Each instance of this class represents one classifier.
---------------------------------------------------------------------------------------------------------------------------------------------------------
MCS Version 1.0: A Python implementation of the MCS learning classifier system algorithm implemented for SNP genetic data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

#!/usr/bin/env python

#Import modules
from MCS_Constants import *
from random import *

class MClassifier:    
    def __init__(self,a=None,b=None):
        """ Overloaded constructor see init1() init2() init3() init4()
        Constructs a classifier based on the type of the arguments. """

        self.condition          =[]
        self.action             = 0
        self.fitness            = 0.0

        self.attCombList = []
        self.attCombos = cons.attributeCombos
        for i in range(self.attCombos):
            self.attCombList.append(str(i))

        if isinstance(a,list):
            self.init1(a,b)
        elif isinstance(b,list):
            self.init2(a,b,)
        elif isinstance(a,int) and isinstance(b,int):
            self.init3(a,b)
        elif isinstance(a,MClassifier):
            self.init4(a)
        else:
            raise NameError('MClassifier unknown init params:')

        
    def init1(self,  situation,  act):
        """ Constructs a classifier with matching condition and specified action.
        @param situation The current problem instance/perception.
        @param act The action of the new classifier. """
    	self.createMatchingCondition(situation)
    	self.action = act
    	self.classifierSetVariables()

  
    def init2(self, numberOfActions,  situation):
        """ Construct matching classifier with random action.
        NOT USED IN THE CURRENT IMPLEMENTATION """
        
    	self.createMatchingCondition(situation)
    	self.createRandomAction(numberOfActions)
    	self.classifierSetVariables()

        
    def init3(self, condLength, numberOfActions):
        """ Construct a classifier with random condition and random action.
        @param condLength The length of the condition of the new classifier.
        @param numberOfActions The number of different actions to chose from. """
        
    	self.createRandomCondition(condLength)
    	self.createRandomAction(numberOfActions)
    	self.classifierSetVariables()
  
        
    def init4(self,clOld):
        """  Constructs an identical MClassifier.  However, the experience of the copy is set to 0 and the numerosity 
        is set to 1 since this is indeed a new individual in a population.
        @param clOld The to be copied classifier. """

    	self.condition       = clOld.condition[:] # a copy of the condition not the same reference
    	self.action          = clOld.action
    	self.fitness       = clOld.fitness 

        
    def createRandomCondition(self,condLength):
        """ Creates a condition randomly considering the constant <code>P_dontcare<\code>. """
        
    	condArray = []
    	for i in range(condLength):
    	    if random() < cons.P_dontcare:
    	        condArray.append(cons.dontCare)
    	    else:
                condArray.append(choice(self.attCombList))
    	self.condition = condArray

        
    def createMatchingCondition(self,cond):
        """ Creates a matching condition considering the constant <code>P_dontcare<\code>. """

    	condLength = len(cond)
    	condArray  = []
        for i in range(condLength):
            if random()< cons.P_dontcare:
                condArray.append(cons.dontCare)
    	    else:
    	        condArray.append(cond[i])
    	self.condition = condArray

        
    def createRandomAction(self, numberOfActions):
        """ Creates a random action.
        @param numberOfActions The number of actions to chose from. 
        NOT USED IN CURRENT IMPLEMENTAION """
    	action = int(random()*numberOfActions)

        
    def classifierSetVariables(self):
        """ Sets the initial variables of a new classifier.
        @param setSize The size of the set the classifier is created in.
        @param time The actual number of instances the MCS learned from so far. """
    	self.fitness  = cons.fitnessIni

        
    def match(self, state):
        """ Returns if the classifier matches in the current situation.
        @param state The current situation which can be the current state or problem instance. """
        
    	if len(self.condition) != len(state):
    	    return False
    	
    	for i in range(len(self.condition)):
    	    if self.condition[i] != cons.dontCare and self.condition[i]!=state[i]:
    	        return False;
    	return True


    def onePointCrossover(self,cl):
        """ Applies one point crossover and returns if the classifiers changed.
        @param cl The second classifier for the crossover application. """
        
        changed = False;
        if random() < cons.pX:
            length = len(self.condition)
            sep1   = int(random()*(length)) +1 

            cond1 = self.condition[:]
            cond2 = cl.condition
            
            if cond1[sep1:] == cond2[sep1:]:
                changed = True
            
            temp = cond1[sep1:] #latter half
            cond1 = cond1[:sep1] + cond2[sep1:]
            cond2 = cond2[:sep1] + temp

            if changed:
                self.condition = cond1
                cl.condition   = cond2

        return changed


    def twoPointCrossover(self,cl):
        """ Applies two point crossover and returns if the classifiers changed.
        @param cl The second classifier for the crossover application. """

    	changed = False;
        if random() < cons.pX:
            # set the two crossover points
    	    length = len(self.condition)
    	    sep1   = int(random()*(length))
    	    sep2   = int(random()*(length)) + 1
    	    if sep1>sep2:
        		help1 = sep1
        		sep1 = sep2
        		sep2 = help1

            elif sep1==sep2:
                sep2= sep2+1
            # Grab the original strings
    	    cond1 = self.condition[:]
    	    cond2 = cl.condition
            
            # switch the middle section of the first and second strings
            for i in range(sep1,sep2):
        		if cond1[i] != cond2[i]:
        		    changed = True
        		    help2   = cond1[i]
        		    cond1[i]= cond2[i]
        		    cond2[i]= help2

    	    if changed:
        		self.condition = cond1
        		cl.condition   = cond2

    	return changed

  
        
    def applyMutation(self, state, numberOfActions):
        """ Applies a niche mutation to the classifier.
        This method calls mutateCondition(state) and mutateAction(numberOfActions) and returns
        if at least one bit or the action was mutated.
        @param state The current situation/problem instance
        @param numberOfActions The maximal number of actions possible in the environment. """
        
        changed = self.mutateCondition(state)
    	if self.mutateAction(numberOfActions):
    	    changed = True
    	return changed


    def mutateCondition(self, state): 
        """ Mutates the condition of the classifier. If one allele is mutated depends on the constant pM.
        This mutation is a niche mutation. It assures that the resulting classifier
        still matches the current situation.
        @param state The current situation/problem instance.  """

    	changed    = False
    	condLength = len(self.condition)
        cond   = self.condition[:]
        
    	for i in range(condLength):
    	    if random()<cons.pM:
                changed= True
                if cond[i]==cons.dontCare:
        		    cond[i] = state[i]
                else:
        		    cond[i] = cons.dontCare

        self.condition = cond
    	return changed

	
    def mutateAction(self, numberOfActions):
        """ Mutates the action of the classifier.
        @param numberOfActions The number of actions/classifications possible in the environment. """

    	changed = False

    	if random()<cons.pM:
            act = int(random()*numberOfActions)
            while act==self.action:
                act = int(random()*numberOfActions)

            self.action = act
    	    changed= True
    	return changed
        
        
    def updateFitness(self, reward, actionSetSize):
        """ Updates the fitness of the classifier according to the relative accuracy.
        @param reward The amount of reward received following an accurate classification
        @param actionSetSize The number of classifiers in the action set. """

    	self.fitness = self.fitness + cons.beta * (reward / actionSetSize - self.fitness)
    	return self.fitness 

            
    def getAction(self):
        """  Returns the action of the classifier. """
    	return self.action
 
 
    def getCondition(self):
        """ Returns the condition of the classifier. """
        return self.condition
        
        
    def getFitness(self):
        """ Returns the fitness of the classifier. """
        return self.fitness

    
    def setFitness(self, fit):
        """  Sets the fitness of the classifier.
        @param fit The new fitness of the classifier. """
    	self.fitness=fit


    def getClassifierGenerality(self):
        """ Calculates and returns the fraction of bits that infer generality (that have a wild card symbol)"""
        
        genCount = 0
        condLength = len(self.condition)
        for x in self.condition:
            if str(x) == cons.dontCare:
                genCount += 1
        classifierGenerality = genCount/float(condLength)
        return classifierGenerality


    def __str__(self):
        """ the string represetation of the classifier
        used for printin purposed (usefull for debugin) """
        return str(self.condition)+"\t"+str(self.action)+"\t"+str(self.fitness)
        

    def printMClassifier(self, pW):
        """ Prints the classifier to the print writer (normally referencing a file).
        The method prints condition action prediction predictionError fitness numerosity experience actionSetSize timeStamp.
        @param pW The writer to which the classifier is written. """
        
        gen = self.getClassifierGenerality()
        for x in self.condition:
            pW.write(x)
        pW.write("\t"+str(self.action)+"\t"+str(self.fitness)+"\t"+str(gen)+"\n") 
