"""
Name: MCS_ClassifierSet.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA 
Note: Code primarily based on python XCS authored by Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
Contact: ryan.j.urbanowicz@darmouth.edu
Created: 3/20/2009
Updated: 10/15/2013
Description: This class handles the different sets of classifiers. It stores each set in an array. The array is initialized to a sufficient large size so 
that no changes in the size of the array will be necessary.
---------------------------------------------------------------------------------------------------------------------------------------------------------
MCS Version 1.0: A Python implementation of the MCS learning classifier system algorithm implemented for SNP genetic data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

#!/usr/bin/env python

#Import modules
from MCS_Classifier import *
from MCS_Environment import *
import sys
from random import *

class MClassifierSet:      
    def __init__(self,a=None,b=None,c=None,d=None):
        """ Creates a new classifier, see init1(), init2() , init3() for more details"""

        self.parentSet=None
        self.clSet=[]
        self.attributeGenList = 0.0
        self.generality = 0.0 
        self.weightedGenerality = 0.0
        
        if isinstance(a,int) and b==None and c==None and d==None:
            self.init1(a)
        elif isinstance(a,list) and isinstance(b,MClassifierSet):
            self.init2(a,b,c,d)
        elif isinstance(a,MClassifierSet):
            self.init3(a,b)
        elif isinstance(a,str) and isinstance(b,list) and isinstance(c,MClassifierSet):
            self.init4(a,b,c)

        
    def init1(self, numberOfActions):
        """ Creates a new, empty population initializing the population array to the maximal population size plus the number of possible actions.
        @param numberOfActions The number of actions possible in the problem. """

        self.parentSet     = None


    def init2(self, state, pop, numberOfActions, maxPopSize):
        """ Constructs a match set out of the population. After the creation, it is checked if the match set covers all possible actions
        in the environment. If one or more actions are not present, covering occurs, generating the missing action(s). If maximal
        population size is reached when covering, deletion occurs.
        @param state The current situation/problem instance.
        @param pop The current population of classifiers.
        @param numberOfActions The number of actions possible in the environment. """

        self.parentSet     = pop
        actionCovered = [False for i in range(numberOfActions)]
        
        for i in range(len(pop.clSet)):
            cl = pop.clSet[i]
            if cl.match(state):
                self.addClassifier(cl)  #Creates a match set
                actionCovered[cl.getAction()] = True # Check to see that all actions are present in the match set

        while True:
            again=False
            for i in range(len(actionCovered)):
                # COVERING
                if not actionCovered[i]:
                    newCl = MClassifier(state, i)
                    self.addClassifier(newCl)
                    pop.addClassifier(newCl)

            # Deletion in the case that covering exceeds pop max. (Altered for MCS which doesn't have numerosity)
            while len(pop.clSet) > maxPopSize:
                cdel = pop.deleteFromPopulation()
                # update the current match set in case a classifier was deleted out of that
                # and redo the loop if now another action is not covered in the match set anymore.
                pos = self.containsClassifier(cdel)
               
                #In case the deleted classifier is in the action set, we might have to generate a new classifier.
                if cdel != None and pos != -1:
            		self.removeClassifier(pos) #deletes from match set
            		if not self.isActionCovered(cdel.getAction()):
            			 again = True
            			 actionCovered[cdel.getAction()] = False
            if not again:
                break
       

    def init3(self, matchSet, action):
        """ Constructs an action set out of the given match set.
        @param matchSet The current match set
        @param action The chosen action for the action set. """
        
        self.parentSet     = matchSet

        for i in range(len(matchSet.clSet)):
            if matchSet.clSet[i].getAction() == action:
                self.addClassifier(matchSet.clSet[i])
        
        
    def init4(self, special, state, pop):
        """ Added to original XCS to allow for a non-destructive match set to be created for evaluation purposes. """
        self.parentSet = pop
               
        for i in range(len(pop.clSet)):
            cl = pop.clSet[i]
            if cl.match(state):
                self.addClassifier(cl)
        
        
    def containsClassifier(self,clP):
        """ Returns the position of the classifier in the set if it is present and -1 otherwise. Modified for python lists. """ 
        
        try:
            return self.clSet.index(clP)
        except ValueError:
            return -1
        
        
    def isActionCovered(self, action):
        """ Returns if the specified action is covered in this set. """
        for cl in self.clSet:
            if cl.getAction() == action:
                return True
        return False

        
    def updateSet(self, reward):
        """ Updates all parameters in the current set (should be the action set).
        @param reward The actual resulting reward after the execution of an action. """
        self.updateFitnessSet(reward)


    def updateFitnessSet(self, reward):
        """ Special function for updating the fitness of the classifiers in the set. """

        for i in range(self.getSize()):
            self.clSet[i].updateFitness(reward, self.getSize())

        
    def runGA(self, state, numberOfActions, maxPopSize, selection):
        """ The Genetic Discovery in MCS takes place here. If a GA takes place, two classifiers are selected
        by roulette wheel selection, possibly crossed and mutated and then inserted.
        @param state  The current situation/problem instance.
        @param numberOfActions The number of actions possible in the environment. """

        # Get random number between 0-1 and if > g than don't perform GA.
        if self.getSize()==0 or random() > cons.gamma:
            return
        
        if selection == 0:
            fitSum =self.getFitnessSum()
            # Select two MClassifiers with roulette Wheel Selection
            cl1P = self.selectMClassifierRW(fitSum)
            cl2P = self.selectMClassifierRW(fitSum)
        else:
            cl1P = self.selectMClassifierT()
            cl2P = self.selectMClassifierT()
  
        cl1  = MClassifier(cl1P)
        cl2  = MClassifier(cl2P)
    
        changed = cl1.onePointCrossover(cl2)
        if changed:
            tempFitness = (cl1.getFitness() + cl2.getFitness())/2.0
            cl1.setFitness(tempFitness)
            cl2.setFitness(tempFitness)
        else:   # ensure that the fitness discount is still applied for having run the GA at all
            cl1.setFitness(cl1.getFitness())
            cl2.setFitness(cl2.getFitness())
            
        cl1.applyMutation(state, numberOfActions)
        cl2.applyMutation(state, numberOfActions)
   
        self.insertDiscoveredMClassifiers(cl1, cl2, cl1P, cl2P, maxPopSize)


    # ROULETTE WHEEL SELECTION    #this implementation has been tested to work correctly.
    def selectMClassifierRW(self,fitSum):
        """ Selects one classifier using roulette wheel selection according to the fitness of the classifiers. """
        
        choiceP = random()*fitSum
        i=0
        sum = self.clSet[i].getFitness()
        while choiceP>sum:
            i=i+1
            sum = sum + self.clSet[i].getFitness()
        return self.clSet[i]


    # TOURNAMENT SELECTION
    def selectMClassifierT(self):
        """  Selects one classifier using tournament selection according to the fitness of the classifiers. """
        
        actionSetSize = len(self.clSet) 
        tSize = int(actionSetSize*cons.theta_Select) # sets the number of items in the action set to be included in the tournament selection
        posList = []
        for i in range(tSize): #hold onto a list of random positions, then select the position with the highest fitness
            pos = randrange(actionSetSize)
            if pos in posList: # make sure that pos is a pos that has not yet been selected.
                pos = randrange(actionSetSize)
            else:
                posList.append(pos)
                
        bestF = 0
        bestC = 0
        for j in posList:
            if self.clSet[j].getFitness() > bestF:
                bestF = self.clSet[j].getFitness()
                bestC = j

        return self.clSet[bestC]

        
    def insertDiscoveredMClassifiers(self,cl1, cl2, cl1P, cl2P, maxPopSize):
        """ Inserts both discovered classifiers keeping the maximal size of the population and possibly doing GA subsumption.
        @param cl1 The first classifier generated by the GA.
        @param cl2 The second classifier generated by the GA.
        @param cl1P The first parent of the two new classifiers.
        @param cl2P The second classifier of the two new classifiers. """
        
        pop = self
        while pop.parentSet!=None: #This works in Python exactly as in Java
            pop = pop.parentSet;

        pop.addMClassifierToPopulation(cl1)
        pop.addMClassifierToPopulation(cl2)

        while pop.getSize() > maxPopSize:
            pop.deleteFromPopulation()


    def addMClassifierToPopulation(self,cl):
        """ Adds the classifier to the population
        @param cl The to be added classifier. """

        pop=self  # set pop to the actual population
        while pop.parentSet!=None:
            pop = pop.parentSet
    
        pop.addClassifier(cl)

        
    def deleteFromPopulation(self): 
        """ Deletes one classifier in the population.  The classifier that will be deleted is chosen by roulette wheel selection
        considering the deletion vote. Altered from XCS """
            
        fitSum = self.getFitnessSum()
        choicePoint = fitSum*random()
        sum=0.0
        i = 0
        for cl in self.clSet:
            sum = sum + self.clSet[i].getFitness()
            if sum > choicePoint:
                self.removeClassifier(i)
                return cl
            i = i + 1
        return None

        
    def addClassifier(self, classifier):
        """ Adds a classifier to the set and increases the numerositySum value accordingly.
        @param classifier The to be added classifier. """
        self.clSet.append(classifier)


    def removeClassifier(self,param=None):
        """ overloaded method see removeClassifier1(classifier), removeClassifier2(pos)"""

        # This is -only one- of the various python ways of implementing overloaded methods.
        if isinstance(param,MClassifier):
            self.removeClassifier1(param)
        else:
            self.removeClassifier2(param)
        
        
    def removeClassifier1(self, classifier):
        """ Removes the specified (possible macro-) classifier from the population.
        The function returns true when the classifier was found and removed and false
        otherwise. It does not update the numerosity sum of the set, neither
        recursively remove classifiers in the parent set. This must be done manually where required.
        Hard modified for Pythonic way of doing things with lists. """
        
        try:
            self.clSet.remove(classifier)
            return True
        except ValueError:
            raise
            
        
    def removeClassifier2(self,pos):
        """ Removes the (possible macro-) classifier at the specified array position from the population.
        The function returns true when the classifier was found and removed and false
        otherwise. It does not update the numerosity of the set, neither
        recursively remove classifiers in the parent set. This must be done manually where required.
        Hard modified for Pythonic way of doing things with lists. """
        
        try:
            del self.clSet[pos]
        except IndexError:
            raise
        return True

        
    def getFitnessSum(self):
        """ Returns the sum of the fitnesses of all classifiers in the set. """
        
        sum=0.0
        for cl in self.clSet:
            sum = sum + cl.getFitness()
        return sum

        
    def getSize(self):
        """ Returns the number of macro-classifiers in the set. """
        return len(self.clSet)  # true size
    
    
    def CSAnalysis(self,env):
        """ Characterizes the current classifier set """
        # Creates the attribute Generality List
        bitLength = env.getAttributeLength()
        wildCount = self.characterizePop()
        self.attributeGenList = self.condenseToAttributes(wildCount,bitLength) 
        
        genSum = 0
        fitGenSum = 0
        fitSum = 0       
        for cl in self.clSet:
            genSum = genSum + cl.getClassifierGenerality()
            fitGenSum = fitGenSum + cl.getClassifierGenerality() * cl.getFitness()
            fitSum = fitSum +  cl.getFitness()
            
        self.generality = float(genSum / len(self.clSet))
        self.weightedGenerality = fitGenSum / (fitSum)
        
        
    def characterizePop(self):
        """ Sums up the #'s within all conditions """
        condition = []
        countList = []
        for i in range(len(self.clSet[0].getCondition())):  #Over length of a binary code condition
            countList.append(int(0))   # makes an empty countList
            
        for cl in self.clSet:
            condition = cl.getCondition()
            for j in range(len(condition)):    #for each character string in the condition
                if condition[j] == '#':
                    countList[j] +=1
        return countList
    
    
    def condenseToAttributes(self, tempList, bitLength):
        """ Takes the results of the above method and condenses the values down to a "per attribute" value. """
        temp = 0
        newList = []
        for i in range(len(tempList)):
            if (i+1)%int(bitLength) != 0:  #first run it will be 1
                temp = temp + tempList[i]
            else:
                newList.append(temp + tempList[i])
                temp = 0
        return newList
    
                          
    def printCSEvaluation(self, pW, trainEval, testEval, learnSec, env, CVPartition):
        """ Prints all information on the Classifier Set evaluation to the designated output file object """
        #Segregate into two parts - Learning Characterisitcs and Pop Characteristics
        trainaccuracy = trainEval[0]
        trainNoMatchFrq = trainEval[1]
        
        testaccuracy = testEval[0]
        testNoMatchFrq = testEval[1]
                
        pW.write("Performance Stats:\n")
        pW.write("Training_Accuracy\tTesting_Accuracy\tNoMatchFreq_Train\tNoMatchFreq_Test\tLearningTime(min)\n")
        pW.write(str(trainaccuracy)+"\t")
        if CVPartition > 1:
            pW.write(str(testaccuracy)+"\t")
        else:
            pW.write("NA\t")
        pW.write(str(trainNoMatchFrq) +"\t")            
        if CVPartition > 1:
            pW.write(str(testNoMatchFrq)+"\t")
        else:
            pW.write("NA\t")
        pW.write(str(learnSec/60)+"\n"+"\n")
        
        pW.write("Population Characterization:\n")   
        pW.write("MacroPopSize\tGenerality\tWeightedGenerality\tAverageFitness\n") 
        pW.write(str(len(self.clSet))+"\t"+str(self.generality)+"\t"+str(self.weightedGenerality)+"\t"+str(self.getFitnessSum()/len(self.clSet))+"\n")

        pW.write("WildSum\n")
        headList = env.getHeaderList()
        for i in range(len(headList)-1):    # Added the -1 to get rid of the Class Header
            if i < len(headList)-2:
                pW.write(str(headList[i])+"\t")
            else:
                pW.write(str(headList[i])+"\n")
                
        # Prints out the Gen Count for each attribute
        for i in range(len(self.attributeGenList)):
            if i < len(self.attributeGenList)-1:
                pW.write(str(self.attributeGenList[i])+"\t")
            else:
                pW.write(str(self.attributeGenList[i])+"\n")        
        
        pW.write("Ruleset Population: \n")
        pW.write("Condition\tAction\tFitness\tGenerality\n")
        for cl in self.clSet:
            cl.printMClassifier(pW)
                    
